@@ -2,6 +2,7 @@
  *  * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
  *  *  Motsognir - The mighty gopher server                               *
  *  *  Copyright (C) 2008-2016 Mateusz Viste                              *
+ *  *  http://motsognir.sourceforge.net                                   *
  *  * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
  *
  * ----------------------------------------------------------------------
@@ -46,7 +47,7 @@
 #include "extmap.h"
 
 /* Constants */
-#define pVer "1.0.8"
+#define pVer "1.0.8.1"
 #define pDate "2008-2016"
 #define HOMEPAGE "http://motsognir.sourceforge.net"
 
@@ -520,7 +521,7 @@ static void about(char *version, char *datestring, char *homepage) {
 
 static void sendbackhttperror(int sock, struct MotsognirConfig *config) {
   char txtline[1024], portstr[16];
-  syslog(LOG_INFO, "HTTP request detected - a HTTP error message is returned.");
+  syslog(LOG_INFO, "HTTP request detected - a HTTP error message is returned");
   sendline(sock, "HTTP/1.1 400 Bad request");
   sendline(sock, "Content-Type: text/html; charset=UTF-8");
   sendline(sock, "Server: Motsognir");
@@ -556,6 +557,20 @@ static void sendbackhttperror(int sock, struct MotsognirConfig *config) {
 }
 
 
+static void sendbackgopherplushack(int sock, struct MotsognirConfig *config) {
+  char txtline[1024];
+  syslog(LOG_INFO, "GOPHER+ request detected - a gopher+ fake redirector is returned");
+  sendline(sock, "+-1");
+  snprintf(txtline, sizeof(txtline), "+INFO: 1Main menu (non-gopher+)\t\t%s\t%d", config->gopherhostname, config->gopherport);
+  sendline(sock, txtline);
+  sendline(sock, "+VIEWS:");
+  sendline(sock, " application/gopher+-menu: <512b>");
+  sendline(sock, "+ABSTRACT:");
+  sendline(sock, " This gopher supports standard gopher access only.");
+  sendline(sock, ".");
+}
+
+
 /* explode a list of directories separated by ':' into a malloc()ed, NULL-terminated array of strings */
 static char **explode_dirlist(char *s) {
   char **res = NULL;
@@ -1458,6 +1473,14 @@ static int requestlookslikehttp(char *req) {
 }
 
 
+/* Looks at a requests to detect whether it is a gopher+ requets for the root directory.
+ * Returns 0 if not, non-zero otherwise. */
+static int requestlookslikegopherplus(char *req) {
+  if ((req[0] == '\t') && (req[1] == '$') && (req[2] == 0)) return(1);
+  return(0);
+}
+
+
 /* check for file evasion. Returns 0 if all is ok. non-zero otherwise. */
 static int checkforevasion(char *gopherroot, char **pubdirlist, char *localfile) {
   char resolvedpath[PATH_MAX];
@@ -1631,6 +1654,17 @@ int main(int argc, char **argv) {
     return(0);
   }
 
+  /* detect gopher+ requests for the root resource, as forged by the UMN gopher
+   * client. this needs to be handled because of a bug in the gopher client,
+   * which makes it output an error instead of fallbacking to standard gopher */
+  if (requestlookslikegopherplus(directorytolist) != 0) {
+    char discarddatabuff[4096];
+    sendbackgopherplushack(sock, &config);
+    recv(sock, discarddatabuff, 4096, MSG_DONTWAIT); /* read whatever request the peer sent us, to drain the socket before closing it (otherwise the tcp stack would trigger a ugly RST) */
+    close(sock);
+    return(0);
+  }
+
   /* detect requests for foreign URLs and return a simple html redirecting page */
   if ((directorytolist[0] == 'U') && (directorytolist[1] == 'R') && (directorytolist[2] == 'L') && (directorytolist[3] == ':')) {
     exturlredirector(sock, directorytolist);
